// src/backend/dynamic_process.cc - Copyright 2005, 2006, University
//               of Padova, dept. of Pure and Applied
//               Mathematics
//
// This file is part of SGPEMv2.
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// SGPEMv2 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SGPEMv2; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

#include "dynamic_process.hh"
#include "static_process.hh"
#include "dynamic_thread.hh"
#include <cassert>

using namespace sgpem;
using std::vector;

DynamicProcess::DynamicProcess(StaticProcess* core) :
  DynamicSchedulable(*core)
{
}
  
DynamicProcess::DynamicProcess(const DynamicProcess &other) :
  Schedulable(), DynamicSchedulable(other), Process()
{
  typedef vector<DynamicThread*>::const_iterator ThreadIt;
  
  const vector<DynamicThread*>& other_threads = other._dynamic_threads;

  for(ThreadIt it = other_threads.begin(); it != other_threads.end(); ++it)
    _dynamic_threads.push_back(new DynamicThread(*(*it)));
}

std::vector<Thread*> 
DynamicProcess::get_threads()
{
  return vector<Thread*>(_dynamic_threads.begin(), _dynamic_threads.end());
}
  
Schedulable::state 
DynamicProcess::get_state() const
{
  typedef vector<DynamicThread*>::const_iterator ThreadIt;
  
  assert(_dynamic_threads.size() > 0);
  
  state result = _dynamic_threads[0]->get_state(); 
  bool no_hole = false;
  
  for(ThreadIt it = _dynamic_threads.begin(); it != _dynamic_threads.end(); ++it)
  {
    // This is the logic behind the code:
    // If there is at least one running or ready thread, the result is 
    // the same. If not, it may be either blocked, future or terminated.
    // This is guessed from the state of the first thread.
    // We have three cases:
    // - in case the first thread is blocked, this will be the result.
    // - in case the first thread is terminated, a blocked thread will 
    //   override the result and make the process blocked too.
    // - in case the first thread is future, a blocked or terminated thread 
    //   will override the result. If this isn't the case, also the absence of at least
    //   one future thread starting at instant 0 will make the result equal to terminated.
    // TODO Is this OK? Must be tested...
    
    if((*it)->get_state() == state_running || (*it)->get_state() == state_ready)
      return (*it)->get_state();
    else if((*it)->get_state() == state_blocked)
      result = state_blocked;
    else if(result == state_future)
    {
      if((*it)->get_state() == state_terminated)
	result = state_terminated;
      // NOTE this is not necessary, an else will equally work, but this is more clear...
      else if((*it)->get_state() == state_future)
	no_hole = no_hole || (*it)->get_arrival_time() == 0;
    }
  }

  if(result == state_future && no_hole == false)
    result = state_terminated;

  return result;
}

void 
DynamicProcess::remove_thread(Thread* thread)
{
  assert(thread != NULL);

  vector<DynamicThread*>::iterator it;
  
  it = std::find(_dynamic_threads.begin(), _dynamic_threads.end(), thread);

  if(it != _dynamic_threads.end())
  {
    _dynamic_threads.erase(it);
    // FIXME remove me and leave the responsibility for deletion to the caller
    // (which is?)
    delete *it;
  }
  
}

void
DynamicProcess::add_thread(DynamicThread* thread)
{
  assert(thread != NULL);

  _dynamic_threads.push_back(thread);
}

void 
DynamicProcess::serialize(SerializeVisitor& translator) const
{
  //FIXME write this code. I'm predictable, I know
}
    
